{ MANDELM.PAS
 
  Parallel Pascal v. 2.0: Copyright (c) 1988 3L Ltd
  Example program: Mandelbrot set evaluation and display
 
  Master task
  Rev 000  28-Jul-88  JF   Created
  Rev 001  10-Feb-89  JF   Changes for version 2.0.1
 
  The application
  ---------------
 
  The application consists of two tasks:
 
  (1) MANDELM (this file).   This is the master task, and runs in the
      root transputer.
  (2) MANDELW.   This is the worker task, and runs in all the other
      transputers of the net.
 
  The flood configurer, FCONFIG, can be used to produce an executable
  file which will automatically distribute the worker tasks across an
  arbitrary network and route work packets from the master to the
  workers.
 
  It is also possible to run the application in a single transputer.
  This will work automatically if the application is configured using
  FCONFIG.  Alternatively, a static single-transputer configuration
  could be built by hand, using CONFIG.  A suitable configuration file
  may be found in MANDEL.CFG.
 
  As well as various routines from the Parallel Fortran run-time library,
  MANDELM must be linked with the CGA primitives module, CGA.BIN.
  A file MANDELM.LNK is supplied, which may be used to link MANDELM,
  like this:
 
     LINKT @MANDELM.LNK,MANDELM.B4

  This is for the T414; changes to MANDELM.LNK are needed for the T800.
 
  Functions of the tasks
  ----------------------
 
  MANDELM is told by the user which part of the Mandelbrot set to
  evaluate.   It then breaks this up into 100 packets, and sends them
  to the network of MANDELW's.   As the results from each return, they
  displayed on the PC's screen.
 
  The data packet which is sent by MANDELM to MANDELW is a record of the
  type COMMAND, as defined in MANDEL.INC;  the data packet which is sent
  back to MANDELM by MANDELW is a record of the type RESULTS.
 
  Internals of MANDELM
  --------------------
 
  The task contains three threads.
 
  (1) The Main thread.
  This runs in the main program.  First, it initialises the other two
  threads and the integer ICHAN as an internal channel.  Then it goes
  into a loop, once round for each Mandelbrot display.  For each, it
  gets instructions from the user, and then sends to the Send thread
  details of the part of the Mandelbrot set to be evaluated, using the
  ICHAN channel.  It keeps track of completed work by examining
  TALLY_DONE, which is incremented by Receive every time a RESULTS
  packet is displayed; whenever it notices that TALLY_DONE has changed,
  it updates the PC's display; and when TALLY_DONE reaches 100, Main
  knows that the display is complete. 
 
  (2) The Send thread.
  This waits until input arrives through the ICHAN channel, telling
  it what part of the Mandelbrot set to display.   When it arrives,
  this input goes straight into the first parts of the COMMAND data
  packet, which is a record of type COMMAND.  Send then breaks
  the job into 100 small jobs.   For each, it adds the details to the
  COMMAND packet, and uses the net_send function to send it off
  to the network of MANDELW's. Notice that the Send thread does not
  specify WHICH worker task is to do any particular job;  this is
  decided by the network of router tasks.
 
  (3) The Receive thread.
  This simply waits till a packet arrives from the network of MANDELW's
  and then displays it.   Each packet contains all the necessary
  information to display it, so Receive does not need to keep track of
  which packet is which.   Every time it does a display, Receive
  increments TALLY_DONE, so that Main can tell when the whole display
  is complete.
}

PROGRAM MANDELM (Input, Output);

$INCLUDE 'CHAN.INC'
$INCLUDE 'NET.INC'
$INCLUDE 'THREAD.INC'
$INCLUDE 'CGA.INC'
$INCLUDE 'MANDEL.INC'

CONST
   WS_Size       = 10000;
   Packets       = 100;
   X_Increment   = (Max_CGA_LoRes_X+1) DIV 10;
   Y_Increment   = (Max_CGA_Y+1) DIV 10;

TYPE
   T_Num         = 1..3;
   Tally         = 0..Packets;

{ Special IMPORTs for the two versions of THREAD_START we need }
IMPORT PROCEDURE Thread_Start_Send ALIAS 'thread_p_start' (
   PROCEDURE P(X_Inc : CGA_LoRes_X; Y_Inc : CGA_Y);
   VAR WS : UNIV Byte; WSSize, Flags, NArgs : Integer;
   X_Inc : CGA_LoRes_X; Y_Inc : CGA_Y );

IMPORT PROCEDURE Thread_Start_Receive ALIAS 'thread_p_start' (
   PROCEDURE P(VAR T : Tally);
   VAR WS : UNIV Byte; WSSize, Flags, NArgs : Integer;
   VAR T : Tally );

VAR
   { Channel for communication between Main and Send }
   Int_Chan                     : Chan_Type;
   { Global area for Main to write and Receive to read }
   Thresholds                   : ARRAY [T_Num] OF Integer;

 
{ The Send thread runs in this procedure. }
PROCEDURE Send (X_Inc : CGA_LoRes_X; Y_Inc : CGA_Y);
VAR
   X                  : CGA_LoRes_X;
   Y                  : CGA_Y;
   C                  : Command;   
   
BEGIN
   WHILE TRUE DO BEGIN

      { Wait here for instructions from Main thread }
      Chan_In_Message (12, C.Details, Int_Chan);    
      { Send off the packets to be done.   Each includes the top-left
        and bottom-right corners of the area to do.   This both tells
        the worker task what values to generate and identifies the
        RESULTS packet when it arrives in the Receive thread (since
        there's no guarantee that the results will arrive in the
        same order that the commands were sent out.)                  }
      X := 0;
      WHILE X < Max_CGA_LoRes_X DO BEGIN
         C.T.Tlx := X;
         C.T.Brx := X + X_Inc-1;
         Y := 0;
         WHILE Y < Max_CGA_Y DO BEGIN
            C.T.Tly := Y;
            C.T.Bry := Y + Y_Inc-1;
            { Send off the command packet }
            Net_Send (Size(C), C, True);
            Y := Y + Y_Inc
         END;
         X := X + X_Inc
      END

   END
END;


{ The Receive thread runs in this procedure }
PROCEDURE Receive (VAR Tally_Done : Tally);
VAR
   Len, I, N                      : Integer;
   Complete                       : Boolean;
   X                              : CGA_LoRes_X;
   Y                              : CGA_Y;
   Col                            : Colour;
   R                              : Results;
   Thr                            : T_Num;

BEGIN
   WHILE TRUE DO BEGIN

      Len := Net_Receive (R, Complete);
      {  The RESULTS packet includes the coordinates of the top-left
         and bottom-right corners of the data, so we know where to
         display them.                                              }
      I := 1;
      FOR Y := R.T.Tly TO R.T.Bry DO BEGIN
         FOR X := R.T.Tlx TO R.T.Brx DO BEGIN 
            N := R.Counts[I]; I := I+1;
            { Received 0 means 1; 255 means 256; etc }
            N := N + 1;
            { Work out colour from N and thresholds }
            Col := 0;
            FOR Thr := 1 TO 3 DO
               IF N > Thresholds[Thr] THEN Col := Thr;
            { Display the dot }
            CGA_LoRes_Plot (X, Y, Col)
         END
      END;
      { Increment tally of packets displayed }
      Tally_Done := Tally_Done + 1;

   END
END;


{ The Main thread runs in this procedure }
PROCEDURE Main;
VAR
   { Workspace for Send and Receive threads }
   Send_WS, Receive_WS              : ARRAY [1..WS_Size] OF Byte;
   User_Input                       : Set_Details;
   T                                : T_Num;
   Range                            : ShortReal;
   Tally_Done, Previous_Tally       : Tally;

BEGIN

   { Text mode, clear screen and sign on }
   Video_Mode (Mono_80col_Text_Mode);
   Writeln ('Parallel Pascal v. 2.0A');
   Writeln ('Example program: Mandelbrot set evaluation and display');
   Writeln ('Copyright (c) 1988 3L Ltd');
   Writeln ('This program requires a Colour Graphics Adaptor (CGA)');
   Writeln;

   { Initialise the channel for communicating with Send }
   Chan_Init (Int_Chan);

   { Start the other threads }
   Tally_Done := 0;
   Thread_Start_Send (Send, Send_WS, WS_Size, Thread_Urgent,
                 2, X_Increment, Y_Increment);
   Thread_Start_Receive (Receive, Receive_WS, WS_Size, Thread_Urgent,
                 1, Tally_Done);

   { Main Loop }
   WHILE TRUE DO BEGIN

      { Ensure that no other threads are using the Run-Time Library
        (in fact, in this case they won't be, but I have done it here
        as an example...                                              }
      Thread_Use_RTL;

      Write ('Input X coordinate:');  Readln (User_Input.X_Coord);
      Write ('Input Y Coordinate:');  Readln (User_Input.Y_Coord);
      Write ('Input Y range:');       Readln (Range);            
      User_Input.Gap := Range / (Max_CGA_Y+1);
      User_Input.Y_Coord := User_Input.Y_Coord + Range;

      FOR T := 1 TO 3 DO BEGIN
         Write ('Threshold', T:2,':');
         Readln (Thresholds[T])
      END;

      { Release Run-Time Library }
      Thread_Free_RTL;

      { CGA low-resolution graphics }
      Video_Mode (CGA_LoRes_Graphics_Mode);

      { Reset count of finished packets to zero: the Receive thread
        will count it back up as the packets come in...              }
      Tally_Done := 0;

      { Send the instructions to Send }
      Chan_Out_Message (12, User_Input, Int_Chan);

      { Until all the packets have been done, just keep updating the
        display when necessary                                       }
      Previous_Tally := 0;
      WHILE Tally_Done < Packets DO BEGIN
         WHILE Tally_Done = Previous_Tally DO Thread_Deschedule;
         CGA_Update;
         Previous_Tally := Tally_Done;
      END;

      { Extra display, in case... }
      CGA_Update;

      { Once again, wait for Run_Time Library; then beep and wait till
        the user hits return                                           }
      Thread_Use_RTL;
      Write (Chr(7));
      Readln;
      Thread_Free_RTL;

      { Clear the screen and text mode again }
      Video_Mode (Mono_80col_Text_Mode);

   END
END;


{ Start here: immediately invoke Main }
BEGIN
   Main
END.
